void audio_callback(void *userdata, Uint8 *stream, int length)
{
  u32 sample_length = length / 2;
  u32 _length;
  s16 *stream_base = (s16 *)stream;
  s32 *source;
  s32 current_sample;

  u32 i;

  SDL_LockMutex(audio.mutex);

  while(((audio.buffer_index - audio.buffer_base) %
   AUDIO_BUFFER_SIZE) < length)
  {
    // Pump remaining cycles if you can.
    SDL_CondWait(audio.cv, audio.mutex);
  }

  if(config.enable_sound)
  {
    if((audio.buffer_base + sample_length) >= AUDIO_BUFFER_SIZE)
    {
      u32 partial_length = (AUDIO_BUFFER_SIZE - audio.buffer_base) * 2;
      sound_copy(audio.buffer_base, partial_length, normal);
      sound_copy(0, length - partial_length, normal);
      audio.buffer_base = (length - partial_length) / 2;
    }
    else
    {
      sound_copy(audio.buffer_base, length, normal);
      audio.buffer_base += sample_length;
    }
  }
  else
  {
    if((audio.buffer_base + sample_length) >= AUDIO_BUFFER_SIZE)
    {
      u32 partial_length = (AUDIO_BUFFER_SIZE - audio.buffer_base) * 2;
      audio.buffer_base = (length - partial_length) / 2;
    }
    else
    {
      audio.buffer_base += sample_length;
    }
    memset(stream, length, 0);
  }

  if(config.fast_forward == 0)
    SDL_CondSignal(audio.cv);

  SDL_UnlockMutex(audio.mutex);
}

void initialize_audio()
{
  printf("initializing audio with frequency %d\n",
   config.audio_output_frequency);

  switch(config.audio_output_frequency)
  {
    case 8000:
    case 11025:
      audio.playback_buffer_size = 512;
      break;

    case 16000:
    case 22050:
      audio.playback_buffer_size = 1024;
      break;

    case 44100:
      audio.playback_buffer_size = 2048;
      break;

    default:
      audio.playback_buffer_size = 4096;
      break;
  }

  audio.playback_buffer_size = 8192;

  SDL_AudioSpec desired_spec =
  {
    config.audio_output_frequency, AUDIO_S16, 2, 0,
    audio.playback_buffer_size / 4, 0, 0, audio_callback, NULL
  };

  SDL_AudioSpec audio_settings;
  SDL_OpenAudio(&desired_spec, &audio_settings);

  audio.output_frequency = audio_settings.freq;
  audio.mutex = SDL_CreateMutex();
  audio.cv = SDL_CreateCond();
}

void audio_exit()
{
  audio.buffer_index = AUDIO_BUFFER_SIZE - 1;
  audio.buffer_base = 0;
  SDL_CondSignal(audio.cv);
  SDL_CloseAudio();
}

// Do not do either of these two without first locking/unlocking audio
// (see functions below)

void audio_signal_callback()
{
  SDL_CondSignal(audio.cv);
}

void audio_unstall_callback()
{
  audio.buffer_index = AUDIO_BUFFER_SIZE - 1;
  audio.buffer_base = 0;
  audio_signal_callback();
}

void audio_wait_callback()
{
  if((((audio.buffer_index - audio.buffer_base) %
   AUDIO_BUFFER_SIZE) > (audio.playback_buffer_size * 3 / 2)) &&
   (config.fast_forward == 0))
  {
    while(((audio.buffer_index - audio.buffer_base) %
     AUDIO_BUFFER_SIZE) > (audio.playback_buffer_size * 3 / 2))
    {
      SDL_CondWait(audio.cv, audio.mutex);
    }
  }
}

void audio_reset_buffer()
{
  audio.buffer_index = 0;
  audio.buffer_base = 0;
  audio_signal_callback();
}

void audio_lock()
{
  SDL_LockMutex(audio.mutex);
}

void audio_unlock()
{
  SDL_UnlockMutex(audio.mutex);
}

void audio_pause()
{
  SDL_PauseAudio(1);
}

void audio_unpause()
{
  SDL_PauseAudio(0);
}

