u32 key_map(u32 keys)
{
  io_buttons_enum new_buttons = IO_BUTTON_NONE;

  switch(keys)
  {
    case SDLK_UP:
      new_buttons |= IO_BUTTON_UP;
      break;

    case SDLK_DOWN:
      new_buttons |= IO_BUTTON_DOWN;
      break;

    case SDLK_LEFT:
      new_buttons |= IO_BUTTON_LEFT;
      break;

    case SDLK_RIGHT:
      new_buttons |= IO_BUTTON_RIGHT;
      break;

    case SDLK_d:
    case SDLK_SPACE:
      new_buttons |= IO_BUTTON_I;
      break;

    case SDLK_s:
    case SDLK_LCTRL:
      new_buttons |= IO_BUTTON_II;
      break;

    case SDLK_a:
      new_buttons |= IO_BUTTON_III;
      break;

    case SDLK_c:
      new_buttons |= IO_BUTTON_IV;
      break;

    case SDLK_x:
      new_buttons |= IO_BUTTON_V;
      break;

    case SDLK_z:
      new_buttons |= IO_BUTTON_VI;
      break;

    case SDLK_RETURN:
      new_buttons |= IO_BUTTON_RUN;
      break;

    case SDLK_RSHIFT:
      new_buttons |= IO_BUTTON_SELECT;
      break;
  }

  return new_buttons;
}

u32 joy_button_map(u32 button)
{
  button_action(button + 4, 0);
}

u32 joy_hat_map(u32 hat_value)
{
  switch(hat_value)
  {
    case SDL_HAT_UP:
      return config_to_io_map[config.pad[PC_BUTTON_UP]];

    case SDL_HAT_RIGHT:
      return config_to_io_map[config.pad[PC_BUTTON_RIGHT]];

    case SDL_HAT_DOWN:
      return config_to_io_map[config.pad[PC_BUTTON_DOWN]];

    case SDL_HAT_LEFT:
      return config_to_io_map[config.pad[PC_BUTTON_LEFT]];

    case SDL_HAT_RIGHTUP:
      return config_to_io_map[config.pad[PC_BUTTON_RIGHT]] |
       config_to_io_map[config.pad[PC_BUTTON_UP]];

    case SDL_HAT_RIGHTDOWN:
      return config_to_io_map[config.pad[PC_BUTTON_RIGHT]] |
       config_to_io_map[config.pad[PC_BUTTON_DOWN]];

    case SDL_HAT_LEFTUP:
      return config_to_io_map[config.pad[PC_BUTTON_LEFT]] |
       config_to_io_map[config.pad[PC_BUTTON_UP]];

    case SDL_HAT_LEFTDOWN:
      return config_to_io_map[config.pad[PC_BUTTON_LEFT]] |
       config_to_io_map[config.pad[PC_BUTTON_DOWN]];

    case SDL_HAT_CENTERED:
      return IO_BUTTON_NONE;
  }
}

void display_debug(u32 type, s32 palette_number)
{
  switch(type)
  {
    case SDLK_F2:
      display_tileset(palette_number);
      break;

    case SDLK_F3:
      display_tilemap(palette_number);
      break;

    case SDLK_F4:
      display_patternset(palette_number);
      break;
  }
  update_screen();
}

void update_input()
{
  SDL_Event event;

  while(SDL_PollEvent(&event))
  {
    switch(event.type)
    {
      case SDL_QUIT:
        quit();

      case SDL_KEYDOWN:
      {
        switch(event.key.keysym.sym)
        {
          case SDLK_ESCAPE:
            quit();

          case SDLK_1:
            config.bg_off ^= 1;
            break;

          case SDLK_2:
            config.spr_off ^= 1;
            break;

          case SDLK_F1:
            set_debug_mode(DEBUG_STEP);
            break;

          case SDLK_F2:
          case SDLK_F3:
          case SDLK_F4:
          {
            s32 palette_number = 0;
            u32 done = 0;
            u32 type = event.key.keysym.sym;
            s32 low = 0;
            u32 high = 15;
            u32 i;

            if(type == SDLK_F3)
            {
              palette_number = -1;
              low = -1;
            }

            audio_pause();

            display_debug(type, palette_number);

            while(done == 0)
            {

              SDL_WaitEvent(&event);
              if(event.type == SDL_KEYDOWN)
              {
                switch(event.key.keysym.sym)
                {
                  case SDLK_UP:
                  case SDLK_LEFT:
                    if(palette_number == low)
                      palette_number = high;
                    else
                      palette_number--;

                    display_debug(type, palette_number);
                    break;

                  case SDLK_DOWN:
                  case SDLK_RIGHT:
                    if(palette_number == high)
                      palette_number = low;
                    else
                      palette_number++;

                    if(palette_number == high)
                      palette_number = low + 1;

                    display_debug(type, palette_number);
                    break;

                  default:
                    done = 1;
                    break;
                }
              }
            }

            set_screen_resolution(320, 240);
            for(i = 0; i < 240; i++)
            {
              vdc.scanline_widths[i] = 320;
            }

            audio_unpause();

            break;
          }

          case SDLK_BACKQUOTE:
            button_action_direct(CONFIG_BUTTON_FAST_FORWARD, 0);
            break;

          case SDLK_F5:
            button_action_direct(CONFIG_BUTTON_SAVE_STATE, 0);
            break;

          case SDLK_F7:
            button_action_direct(CONFIG_BUTTON_LOAD_STATE, 0);
            break;

          case SDLK_m:
          case SDLK_BACKSPACE:
            button_action_direct(CONFIG_BUTTON_MENU, 0);
            break;

          default:
            io.button_status &= ~(key_map(event.key.keysym.sym));
            break;
        }
        break;
      }

      case SDL_KEYUP:
      {
        io.button_status |= key_map(event.key.keysym.sym);
        break;
      }

      case SDL_JOYBUTTONDOWN:
      {
        io.button_status &= ~(joy_button_map(event.jbutton.button));
        break;
      }

      case SDL_JOYBUTTONUP:
      {
        io.button_status |= joy_button_map(event.jbutton.button);
        break;
      }

      case SDL_JOYHATMOTION:
      {
        io.button_status |=
         IO_BUTTON_UP | IO_BUTTON_DOWN | IO_BUTTON_LEFT | IO_BUTTON_RIGHT;
        io.button_status &= ~(joy_hat_map(event.jhat.value));
        break;
      }
    }
  }

  rapid_counter ^= 1;
}

u32 platform_specific_button_action_direct(u32 button, u32 old_action)
{
}

u8 gui_actions[16];

void clear_gui_actions()
{
  memset(gui_actions, 0, 16);
}

gui_action_type joy_map_gui_action(u32 button)
{
  switch(button)
  {
    case 1:
      return CURSOR_EXIT;

    case 2:
      return CURSOR_SELECT;

    case 0:
      return CURSOR_BACK;

    case 4:
      return CURSOR_PAGE_UP;

    case 5:
      return CURSOR_PAGE_DOWN;
  }

  return CURSOR_NONE;
}

gui_action_type joy_hat_map_gui_action(u32 hat_value)
{
  switch(hat_value)
  {
    case SDL_HAT_UP:
      return CURSOR_UP;

    case SDL_HAT_RIGHT:
      return CURSOR_RIGHT;

    case SDL_HAT_DOWN:
      return CURSOR_DOWN;

    case SDL_HAT_LEFT:
      return CURSOR_LEFT;

    case SDL_HAT_CENTERED:
      clear_gui_actions();
      break;
  }

  return CURSOR_NONE;
}

gui_action_type key_map_gui_action(u32 key)
{
  switch(key)
  {
    case SDLK_ESCAPE:
      return CURSOR_EXIT;

    case SDLK_DOWN:
      return CURSOR_DOWN;

    case SDLK_UP:
      return CURSOR_UP;

    case SDLK_LEFT:
      return CURSOR_LEFT;

    case SDLK_RIGHT:
      return CURSOR_RIGHT;

    case SDLK_RETURN:
      return CURSOR_SELECT;

    case SDLK_BACKSPACE:
      return CURSOR_BACK;

    case SDLK_PAGEUP:
      return CURSOR_PAGE_UP;

    case SDLK_PAGEDOWN:
      return CURSOR_PAGE_DOWN;
  }

  return CURSOR_NONE;
}

gui_action_type get_gui_input()
{
  SDL_Event event;
  gui_action_type gui_action = CURSOR_NONE;

  static u64 button_repeat_timestamp;
  static button_repeat_state_type button_repeat_state = BUTTON_NOT_HELD;
  static u32 button_repeat = 0;
  static gui_action_type cursor_repeat = CURSOR_NONE;

  delay_us(10000);

  while(SDL_PollEvent(&event))
  {
    switch(event.type)
    {
      case SDL_QUIT:
        quit();

      case SDL_KEYDOWN:
        gui_action = key_map_gui_action(event.key.keysym.sym);
        gui_actions[gui_action] = 1;
        break;

      case SDL_KEYUP:
        gui_actions[key_map_gui_action(event.key.keysym.sym)] = 0;
        break;

      case SDL_JOYBUTTONDOWN:
        gui_action = joy_map_gui_action(event.jbutton.button);
        gui_actions[gui_action] = 1;
        break;

      case SDL_JOYBUTTONUP:
        gui_actions[joy_map_gui_action(event.jbutton.button)] = 0;
        break;

      case SDL_JOYHATMOTION:
        gui_action = joy_hat_map_gui_action(event.jhat.value);
        gui_actions[gui_action] = 1;
        break;
    }
  }


  if(gui_action != CURSOR_NONE)
  {
    get_ticks_us(&button_repeat_timestamp);
    button_repeat_state = BUTTON_HELD_INITIAL;
    cursor_repeat = gui_action;
  }
  else
  {
    if(gui_actions[cursor_repeat])
    {
      u64 new_ticks;
      get_ticks_us(&new_ticks);

      if(button_repeat_state == BUTTON_HELD_INITIAL)
      {
        if((new_ticks - button_repeat_timestamp) >
         BUTTON_REPEAT_START)
        {
          gui_action = cursor_repeat;
          button_repeat_timestamp = new_ticks;
          button_repeat_state = BUTTON_HELD_REPEAT;
        }
      }

      if(button_repeat_state == BUTTON_HELD_REPEAT)
      {
        if((new_ticks - button_repeat_timestamp) >
         BUTTON_REPEAT_CONTINUE)
        {
          gui_action = cursor_repeat;
          button_repeat_timestamp = new_ticks;
        }
      }
    }
  }

  return gui_action;
}

void initialize_event()
{
  u32 joystick_count = SDL_NumJoysticks();

  if(joystick_count > 0)
  {
    SDL_JoystickOpen(0);
    SDL_JoystickEventState(SDL_ENABLE);
  }
}

