#include "common.h"

// Make menu.c match this.

u32 config_to_io_map[] =
{
  IO_BUTTON_UP,
  IO_BUTTON_DOWN,
  IO_BUTTON_LEFT,
  IO_BUTTON_RIGHT,
  IO_BUTTON_I,
  IO_BUTTON_II,
  IO_BUTTON_III,
  IO_BUTTON_IV,
  IO_BUTTON_V,
  IO_BUTTON_VI,
  IO_BUTTON_RUN,
  IO_BUTTON_SELECT
};

// A config button has been pressed pressed, perform the action and return
// the button too.

u32 rapid_counter = 0;

u32 platform_specific_button_action_direct(u32 button, u32 old_action);

u32 button_action_direct(u32 button, u32 old_action)
{
  u32 io_buttons = IO_BUTTON_NONE;

  if(button <= CONFIG_BUTTON_SELECT)
  {
    io_buttons = config_to_io_map[button];
  }
  else

  switch(button)
  {
    case CONFIG_BUTTON_RAPID_I:
      if(rapid_counter)
        io_buttons = IO_BUTTON_I;
      break;

    case CONFIG_BUTTON_RAPID_II:
      if(rapid_counter)
        io_buttons = IO_BUTTON_II;
      break;

    case CONFIG_BUTTON_RAPID_III:
      if(rapid_counter)
        io_buttons = IO_BUTTON_III;
      break;

    case CONFIG_BUTTON_RAPID_IV:
      if(rapid_counter)
        io_buttons = IO_BUTTON_IV;
      break;

    case CONFIG_BUTTON_RAPID_V:
      if(rapid_counter)
        io_buttons = IO_BUTTON_V;
      break;

    case CONFIG_BUTTON_RAPID_VI:
      if(rapid_counter)
        io_buttons = IO_BUTTON_VI;
      break;

    case CONFIG_BUTTON_MENU:
      menu(0);
      io_buttons = IO_BUTTON_CLEAR;
      break;

    case CONFIG_BUTTON_LOAD_STATE:
      if(!old_action)
      {
        char state_name[MAX_PATH];
        sprintf(state_name, "%s_%d.svs", config.rom_filename,
         config.savestate_number);
        load_state(state_name);
      }
      break;

    case CONFIG_BUTTON_SAVE_STATE:
      if(!old_action)
      {
        char state_name[MAX_PATH];
        sprintf(state_name, "%s_%d.svs", config.rom_filename,
         config.savestate_number);

        if((config.snapshot_format) == SS_SNAPSHOT_OFF)
        {
          save_state(state_name, NULL);
        }
        else
        {
          u16 *screen_capture = malloc(320 * 240 * sizeof(u16));
          copy_screen(screen_capture);
          save_state(state_name, screen_capture);
          free(screen_capture);
        }
      }
      break;

    case CONFIG_BUTTON_FAST_FORWARD:
      if(!old_action)
        config.fast_forward ^= 1;

      break;

    default:
      io_buttons |= platform_specific_button_action_direct(button, old_action);
      break;
  }

  return io_buttons;
}

u32 button_action(u32 button, u32 old_action)
{
  // Find out which action it corresponds to
  return button_action_direct(config.pad[button], old_action);
}

#ifdef SDL_EVENT
  #include "SDL/SDL_event.c"
#endif

#ifdef GP2X_EVENT
  #include "gp2x/gp2x_event.c"
#endif

#ifdef WIZ_EVENT
  #include "wiz/wiz_event.c"
#endif

#ifdef PSP_EVENT
  #include "psp/psp_event.c"
#endif

