#include "common.h"

#ifdef SDL_MAIN
  #include "SDL/SDL_main.c"
#endif

#ifdef GP2X_MAIN
  #include "gp2x/gp2x_main.c"
#endif

#ifdef WIZ_MAIN
  #include "wiz/wiz_main.c"
#endif

#ifdef PSP_MAIN
  #include "psp/psp_main.c"
#endif


#ifdef PSP_BUILD

PSP_MODULE_INFO("Temper", PSP_MODULE_USER, 1, 1);
PSP_MAIN_THREAD_ATTR(PSP_THREAD_ATTR_USER);

void get_ticks_us(u64 *ticks_return)
{
  u64 ticks;
  sceRtcGetCurrentTick(&ticks);

  *ticks_return = (ticks * 1000000) / sceRtcGetTickResolution();
}

void delay_us(u32 delay)
{
  sceKernelDelayThread(delay);
}

config_struct config =
{
  // u32 pad[16];
  {
    CONFIG_BUTTON_UP, CONFIG_BUTTON_DOWN, CONFIG_BUTTON_LEFT,
    CONFIG_BUTTON_RIGHT, CONFIG_BUTTON_NONE, CONFIG_BUTTON_II,
    CONFIG_BUTTON_I, CONFIG_BUTTON_MENU, CONFIG_BUTTON_NONE,
    CONFIG_BUTTON_NONE, CONFIG_BUTTON_NONE, CONFIG_BUTTON_NONE,
    CONFIG_BUTTON_SELECT, CONFIG_BUTTON_RUN, CONFIG_BUTTON_NONE,
    CONFIG_BUTTON_NONE
  },
  1,                 // u32 show_fps;
  1,                 // u32 enable_sound;
  1,                 // u32 fast_forward;
  44100,             // u32 audio_output_frequency;
  1,                 // u32 patch_idle_loops;
  SS_SNAPSHOT_OFF,   // u32 snapshot_format;
  0,                 // u32 force_usa;

  333,               // u32 clock_speed; start here?
  0,                 // u32 ram_timings;
  100,               // u32 gamma_percent;
  0,                 // u32 six_button_pad;
  CD_SYSTEM_TYPE_V3, // cd_system_type_enum cd_system_type;
  1,                 // u32 bz2_savestates;
  0,                 // u32 per_game_bram;
  0,                 // u32 sound_double;
  PSP_SCALE_NONE,    // u32 scale_factor;
  0,                 // u32 fast_cd_access;
  0,                 // u32 fast_cd_load;
  0,                 // u32 scale_width;
  0,                 // u32 unlimit_sprites
  0,                 // u32 compatibility_mode
};

#endif


void setup_main_dirs()
{
  struct stat sb;

  getcwd(config.main_path, MAX_PATH);

  if(stat("save_states", &sb))
  {
    printf("save_states directory doesn't exist, creating\n");
    make_directory("save_states");
  }

  if(stat("config", &sb))
  {
    printf("config directory doesn't exist, creating\n");
    make_directory("config");
  }

  if(stat("bram", &sb))
  {
    printf("bram directory doesn't exist, creating\n");
    make_directory("bram");
  }

  if(stat("images", &sb))
  {
    printf("images directory doesn't exist, creating\n");
    make_directory("images");
  }
}

const u32 benchmark_frame_interval = 1000;

void benchmark_step()
{
  static u64 benchmark_ms;
  static u32 benchmark_number = 0;

#ifdef PSP_BUILD
  static PspDebugProfilerRegs *profile_struct;
#endif

  switch(benchmark_number)
  {
    case 0:
#ifdef PSP_BUILD
      profile_struct = sceKernelReferGlobalProfiler();

      memset(profile_struct, 0, sizeof(PspDebugProfilerRegs));
      profile_struct->enable = 1;
#endif
      delay_us(1000 * 2500);

      get_ticks_us(&benchmark_ms);
      break;

    case 1:
    {
      u64 new_ms;
      u32 ms_delta;

      get_ticks_us(&new_ms);
      ms_delta = (new_ms - benchmark_ms) / 1000;

      printf("benchmark took %d ms (%lf ms per frame)\n",
       ms_delta, (double)ms_delta / benchmark_frame_interval);

#ifdef PSP_BUILD
      {
        u32 total_stalled_cycles;

        profile_struct->enable = 0;

        total_stalled_cycles = profile_struct->internal +
         profile_struct->memory + profile_struct->copz +
         profile_struct->vfpu;

        printf("PSP profile results:\n"
         "system cycles:    %d\n"
         "cpu cycles:       %d\n"
         "stalled cycles:   %d (%d) (%lf%%)\n"
         "  internal:       %d (%lf%%)\n"
         "  memory:         %d (%lf%%)\n"
         "  copz:           %d (%lf%%)\n"
         "  vfpu:           %d (%lf%%)\n"
         "bus accesses:     %d\n"
         "uncached load:    %d\n"
         "uncached store:   %d\n"
         "cached loads:     %d\n"
         "cached stores:    %d\n"
         "icache misses:    %d\n"
         "dcache misses:    %d\n"
         "dcache writeback: %d\n",

         profile_struct->systemck,
         profile_struct->cpuck,
         total_stalled_cycles,
         profile_struct->sleep,
         percent_of(total_stalled_cycles, profile_struct->cpuck),
         profile_struct->internal,
         percent_of(profile_struct->internal, total_stalled_cycles),
         profile_struct->memory,
         percent_of(profile_struct->memory, total_stalled_cycles),
         profile_struct->copz,
         percent_of(profile_struct->copz, total_stalled_cycles),
         profile_struct->vfpu,
         percent_of(profile_struct->vfpu, total_stalled_cycles),
         profile_struct->bus_access,
         profile_struct->uncached_load,
         profile_struct->uncached_store,
         profile_struct->cached_load,
         profile_struct->cached_store,
         profile_struct->i_miss,
         profile_struct->d_miss,
         profile_struct->d_writeback);
      }
#endif
      quit();
      break;
    }
  }

  benchmark_number++;
}

int main(int argc, char *argv[])
{
  u32 benchmark_mode = 0;
  u32 benchmark_frames;
  u32 load_state0 = 0;
  u32 i;

#ifdef PSP_BUILD
  delay_us(2000000);

  scePowerSetClockFrequency(config.clock_speed, config.clock_speed,
   config.clock_speed / 2);
#endif

  platform_initialize();

  // These are offsets needed by ARM ASM files. Uncomment this if they
  // need to be updated.

/*  printf("vdc.screen_width: %x\n", (u32)(&vdc.screen_width) - (u32)(&vdc));
  printf("vdc.status: %x\n", (u32)(&vdc.status) - (u32)(&vdc));
  printf("vdc.cr: %x\n", (u32)(&vdc.cr) - (u32)(&vdc));
  printf("vdc.mwr: %x\n", (u32)(&vdc.mwr) - (u32)(&vdc));
  printf("vdc.bxr: %x\n", (u32)(&vdc.bxr) - (u32)(&vdc));
  printf("vdc.effective_byr: %x\n", (u32)(&vdc.effective_byr) - (u32)(&vdc));
  printf("vdc.tile_cache: %x\n", (u32)(&vdc.tile_cache) - (u32)(&vdc));
  printf("vdc.pattern_cache: %x\n", (u32)(&vdc.pattern_cache) - (u32)(&vdc));
  printf("vdc.vram: %x\n", (u32)(&vdc.vram) - (u32)(&vdc));
  printf("vdc.dirty_tiles: %x\n", (u32)(&vdc.dirty_tiles) - (u32)(&vdc));
  printf("vdc.dirty_patterns: %x\n",
   (u32)(&vdc.dirty_patterns) - (u32)(&vdc));
  printf("vdc.overdraw_width_offset: %x\n",
   (u32)(&vdc.overdraw_width_offset) - (u32)(&vdc));
  printf("memory.mpr_translated: %x\n", (u32)(&memory.mpr_translated) -
   (u32)(&memory));
  printf("sizeof(sat_cache_line_entry_struct) = %x\n",
   sizeof(sat_cache_line_entry_struct));
  printf("sizeof(sat_cache_line_struct) = %x\n",
   sizeof(sat_cache_line_struct));
  printf("cd.work_ram = %x\n", cd.work_ram);
  printf("memory.work_ram = %x\n", memory.work_ram);

  printf("debug.mode: %x\n", (u32)(&debug.mode) - (u32)(&debug));
  printf("debug.breakpoint: %x\n", (u32)(&debug.breakpoint) - (u32)(&debug));
  printf("debug.instruction_count: %x\n", (u32)(&debug.instruction_count)
   - (u32)(&debug));
  printf("debug.breakpoint_original: %x\n", (u32)(&debug.breakpoint_original)
   - (u32)(&debug)); */

  printf("Setting up main directories.\n");
  setup_main_dirs();

  printf("Initialize PCE\n");
  initialize_pce();

  printf("Loading secondary configuration file\n");
  load_directory_config_file("temper.cf2");

  if(argc > 1)
  {
    printf("Loading ROM %s\n", argv[1]);

    if(load_rom(argv[1]) == -1)
    {
      printf("Error: Could not load %s\n", argv[1]);
      return -1;
    }
    reset_pce();
  }
  else
  {
    printf("Loading primary configuration file\n");
    load_config_file("temper.cfg");
    printf("Starting menu\n");
    menu(1);
  }

  printf("Processing arguments\n");

  for(i = 0; i < argc; i++)
  {
    if(!strcmp(argv[i], "-d"))
      set_debug_mode(DEBUG_STEP);

    if(!strcmp(argv[i], "-b"))
    {
      benchmark_mode = 1;
      benchmark_frames = benchmark_frame_interval;
      config.fast_forward = 1;
    }

    if(!strcmp(argv[i], "-f"))
      config.fast_forward = 1;

    if(!strcmp(argv[i], "-s"))
      load_state0 = 1;
  }

  if((argc > 2) && (argv[argc - 1][0] != '-'))
  {
    debug.breakpoint = strtol(argv[argc - 1], NULL, 16) + 1;
    debug.breakpoint_original = debug.breakpoint;
    set_debug_mode(DEBUG_COUNTDOWN_BREAKPOINT);
  }

  if(load_state0)
  {
    char state_name[MAX_PATH];
    sprintf(state_name, "%s_0.svs", config.rom_filename);
    //config.fast_forward = 1;

    load_state(state_name);
  }

  if(benchmark_mode)
  {
    char state_name[MAX_PATH];
    sprintf(state_name, "%s_0.svs", config.rom_filename);

    load_state(state_name);

    benchmark_step();
  }

  audio_unpause();

  while(1)
  {

#ifdef FASTFORWARD_FRAMESKIP
    if(config.fast_forward && !benchmark_mode)
    {
      static u32 frameskip_counter = 0;

      update_frame(frameskip_counter == 0);
      frameskip_counter = (frameskip_counter + 1) %
       FASTFORWARD_FRAMESKIP_RATE;
    }
    else
    {
      update_frame(0);
    }

#else
    update_frame(0);
#endif

    update_input();

    audio_sync_start();
    update_psg();
    update_cdda();
    audio_sync_end();

    synchronize();

    if(benchmark_mode)
    {
      benchmark_frames--;
      if(benchmark_frames == 0)
      {
        benchmark_step();
        benchmark_frames = benchmark_frame_interval;
      }
    }
  }

  return 0;
}

#ifdef PSP_BUILD

void synchronize()
{
  static u64 frame_wait = 0;

  frame_wait += 16667;

  if(config.fast_forward == 0)
    delay_us(1);
}

#endif


void initialize_pce()
{
  initialize_video();
  initialize_memory();
  initialize_io();
  initialize_irq();
  initialize_timer();
  initialize_psg();
  initialize_cpu();
  initialize_cd();
  initialize_adpcm();
  initialize_arcade_card();

  printf("init debug\n");
  initialize_debug();
}

void reset_pce()
{
  reset_video();
  reset_memory();
  reset_io();
  reset_irq();
  reset_timer();
  reset_psg();
  reset_cpu();
  reset_cd();
  reset_adpcm();
  reset_arcade_card();

  reset_debug();
}

// For now this will save a full sized snapshot.

#define SAVESTATE_MAX_SIZE ((1024 * 256 * 3) + (1024 * 1024 * 2))

void save_state(char *file_name, u16 *snapshot)
{
  char path[MAX_PATH];
  sprintf(path, "%s%csave_states%c%s", config.main_path, DIR_SEPARATOR_CHAR,
   DIR_SEPARATOR_CHAR, file_name);
  u8 *savestate_buffer = malloc(SAVESTATE_MAX_SIZE);
  u32 snapshot_length = 0;

  printf("saving state to file named %s\n", path);

  file_write_mem_open(savestate_file, path, savestate_buffer);
  savestate_header_struct savestate_header;

  // The header is TMSS for "Temper Savestate"
  savestate_header.magic_string[0] = 'T';
  savestate_header.magic_string[1] = 'M';
  savestate_header.magic_string[2] = 'S';
  savestate_header.magic_string[3] = 'S';

  // Version 1 = first version. 0 is reserved as invalid.
  savestate_header.version = TEMPER_SAVESTATE_VERSION;

  savestate_header.extensions = 0;

  // Only save a snapshot if it's given a valid one.

  if(snapshot != NULL)
  {
    savestate_header.extensions |= SS_EXT_SNAPSHOT;
    savestate_header.snapshot_format = SS_SNAPSHOT_FULL;
  }

  if(config.cd_loaded)
  {
    savestate_header.extensions |= SS_EXT_CDROM_DATA;

    if((config.cd_system_type == CD_SYSTEM_TYPE_V3) ||
     (config.cd_system_type == CD_SYSTEM_TYPE_ACD))
    {
      savestate_header.extensions |= SS_EXT_SUPER_CD_DATA;
    }

    if(config.cd_system_type == CD_SYSTEM_TYPE_ACD)
      savestate_header.extensions |= SS_EXT_ARCADE_CARD_DATA;
  }

  if(memory.sf2_region != -1)
    savestate_header.extensions |= SS_EXT_SF2_DATA;

  if(config.six_button_pad)
    savestate_header.extensions |= SS_EXT_6BUTTON_DATA;

  if(config.populous_loaded)
    savestate_header.extensions |= SS_EXT_POPULOUS_DATA;

  if(config.bz2_savestates)
    savestate_header.extensions |= SS_EXT_BZ2_COMPRESSED;

  file_write_mem_array(savestate_file, savestate_header.magic_string);
  file_write_mem_variable(savestate_file, savestate_header.version);
  file_write_mem_variable(savestate_file, savestate_header.extensions);
  file_write_mem_variable(savestate_file, savestate_header.snapshot_format);
  file_write_mem_array(savestate_file, savestate_header.reserved);

  if(savestate_header.extensions & SS_EXT_SNAPSHOT)
  {
    snapshot_length = 320 * 240 * 2;
    file_write_mem(savestate_file, snapshot, snapshot_length);
  }

  video_store_savestate(savestate_file);
  memory_store_savestate(savestate_file);
  io_store_savestate(savestate_file);
  irq_store_savestate(savestate_file);
  timer_store_savestate(savestate_file);
  psg_store_savestate(savestate_file);
  cpu_store_savestate(savestate_file);

  if(config.cd_loaded)
  {
    cd_store_savestate(savestate_file);
    adpcm_store_savestate(savestate_file);

    if(savestate_header.extensions & SS_EXT_SUPER_CD_DATA)
      file_write_mem_array(savestate_file, cd.ext_ram);

    if(savestate_header.extensions & SS_EXT_ARCADE_CARD_DATA)
      arcade_card_store_savestate(savestate_file);
  }

  if((savestate_header.extensions & SS_EXT_SF2_DATA))
    file_write_mem_variable(savestate_file, memory.sf2_region);

  if((savestate_header.extensions & SS_EXT_6BUTTON_DATA))
    file_write_mem_variable(savestate_file, io.six_button_toggle);

  if((savestate_header.extensions & SS_EXT_POPULOUS_DATA))
    file_write_mem_array(savestate_file, memory.hucard_rom + (0x40 * 0x2000));

#ifdef BZ_SUPPORT
  if(savestate_header.extensions & SS_EXT_BZ2_COMPRESSED)
  {
    int error_code;
    BZFILE *bz_savestate_file;
    u32 extra_length = SAVESTATE_HEADER_LENGTH + snapshot_length;

    file_write_mem_save_fixed(savestate_file, extra_length);
    bz_savestate_file = BZ2_bzWriteOpen(&error_code, savestate_file->_file,
     4, 0, 30);

    printf("saving bz2 compressed portion of state at offset %x\n",
     extra_length);
    BZ2_bzWrite(&error_code, bz_savestate_file, savestate_file->buffer +
     extra_length, (savestate_file->buffer_ptr - savestate_file->buffer) -
     extra_length);
    BZ2_bzWriteClose(&error_code, bz_savestate_file, 0, NULL, NULL);
  }
  else
#endif
  {
    file_write_mem_save(savestate_file);
  }
  file_write_mem_close(savestate_file);

  config.saved_state_counter = 20;
  free(savestate_buffer);
}

savestate_extension_enum load_state(char *file_name)
{
  char path[MAX_PATH];
  sprintf(path, "%s%csave_states%c%s", config.main_path, DIR_SEPARATOR_CHAR,
   DIR_SEPARATOR_CHAR, file_name);
  savestate_header_struct savestate_header;
  u8 *savestate_buffer = malloc(SAVESTATE_MAX_SIZE);

  audio_pause();

  file_read_mem_open(savestate_file, path, savestate_buffer);
  printf("loading state %s\n", path);

  if(file_read_mem_invalid(savestate_file))
  {
    printf("savestate %s does not exist\n", path);
    free(savestate_buffer);
    return SS_EXT_INVALID;
  }

  // Don't want things to hang as soon as the state is loaded because
  // of a deadlock between the audio buffer and the audio generation,
  // so make it wait now.

  file_read_mem_load_fixed(savestate_file, SAVESTATE_HEADER_LENGTH);

  file_read_mem_array(savestate_file, savestate_header.magic_string);
  file_read_mem_variable(savestate_file, savestate_header.version);
  file_read_mem_variable(savestate_file, savestate_header.extensions);
  file_read_mem_variable(savestate_file, savestate_header.snapshot_format);
  file_read_mem_array(savestate_file, savestate_header.reserved);

  // First, the savestate string must match.
  if(strncmp(savestate_header.magic_string, "TMSS", 4))
    goto invalid;

  // Right now this only supports savestate version 1. Later if more
  // are done it'll probably support those too! Version 0 is right out
  // though.

  // If savestates change too much there can be a lower limit here too.
  if(savestate_header.version > TEMPER_SAVESTATE_VERSION)
  {
    printf("Error: Savestate version not supported by Temper.\n");
    goto invalid;
  }

  config.savestate_version = savestate_header.version;

  // Does it have a snapshot? What do we care here? Skip it.
  if(savestate_header.extensions & SS_EXT_SNAPSHOT)
  {
    // Okay, must support these are they're added.
    switch(savestate_header.snapshot_format)
    {
      default:
      case SS_SNAPSHOT_FULL:
        file_read_mem_load_null(savestate_file, 320 * 240 * 2);
        file_read_mem_skip_forward(savestate_file, 320 * 240 * 2);
        break;
    }
  }

  // Now we can load all the actual stuff. In the future the extension
  // fields will dictate loading more, but this is the basic set.

#ifdef BZ_SUPPORT
  if(savestate_header.extensions & SS_EXT_BZ2_COMPRESSED)
  {
    int error_code;

    BZFILE *bz_savestate_file = BZ2_bzReadOpen(&error_code,
     savestate_file->_file, 0, 1, NULL, 0);
    BZ2_bzRead(&error_code, bz_savestate_file, savestate_file->buffer,
     SAVESTATE_MAX_SIZE);
    savestate_file->buffer_ptr = savestate_file->buffer;
    BZ2_bzReadClose(&error_code, bz_savestate_file);
  }
  else
#endif
  {
    file_read_mem_load(savestate_file);
  }

  video_load_savestate(savestate_file);
  memory_load_savestate(savestate_file);
  io_load_savestate(savestate_file);
  irq_load_savestate(savestate_file);
  timer_load_savestate(savestate_file);
  psg_load_savestate(savestate_file);
  cpu_load_savestate(savestate_file);

  config.cd_loaded = 0;
  config.populous_loaded = 0;
  memory.sf2_region = -1;

  if(savestate_header.extensions & SS_EXT_CDROM_DATA)
  {
    config.cd_loaded = 1;
    cd_load_savestate(savestate_file);
    adpcm_load_savestate(savestate_file);

    if(savestate_header.extensions & SS_EXT_SUPER_CD_DATA)
      file_read_mem_array(savestate_file, cd.ext_ram);

    if(savestate_header.extensions & SS_EXT_ARCADE_CARD_DATA)
      arcade_card_load_savestate(savestate_file);
  }

  if(savestate_header.extensions & SS_EXT_SF2_DATA)
  {
    u32 current_sf2_region;
    file_read_mem_variable(savestate_file, current_sf2_region);
    map_sf2(current_sf2_region);
  }

  if(savestate_header.extensions & SS_EXT_6BUTTON_DATA)
    file_read_mem_variable(savestate_file, io.six_button_toggle);

  if(savestate_header.extensions & SS_EXT_POPULOUS_DATA)
  {
    config.populous_loaded = 1;
    map_populous_ram();
    file_read_mem_array(savestate_file, memory.hucard_rom + (0x40 * 0x2000));
  }

  config.loaded_state_counter = 20;


  // And return the extensions we loaded.
  free(savestate_buffer);
  audio_unpause();
  return savestate_header.extensions;

 invalid:
  file_read_mem_close(savestate_file);

  printf("An error occured while trying to load the savestate file.\n");
  free(savestate_buffer);
  audio_unpause();
  return SS_EXT_INVALID;
}

savestate_extension_enum load_state_snapshot(char *file_name, u16 *buffer)
{
  char path[MAX_PATH];
  sprintf(path, "%s%csave_states%c%s", config.main_path, DIR_SEPARATOR_CHAR,
   DIR_SEPARATOR_CHAR, file_name);
  savestate_header_struct savestate_header;
  u8 *savestate_buffer = malloc(SAVESTATE_HEADER_LENGTH + (320 * 240 * 2));

  file_read_mem_open(savestate_file, path, savestate_buffer);

  if(file_read_mem_invalid(savestate_file))
  {
    free(savestate_buffer);
    return SS_EXT_INVALID;
  }

  file_read_mem_load_fixed(savestate_file, SAVESTATE_HEADER_LENGTH);

  file_read_mem_array(savestate_file, savestate_header.magic_string);
  file_read_mem_variable(savestate_file, savestate_header.version);
  file_read_mem_variable(savestate_file, savestate_header.extensions);
  file_read_mem_variable(savestate_file, savestate_header.snapshot_format);
  file_read_mem_array(savestate_file, savestate_header.reserved);

  // First, the savestate string must match.
  if(strncmp(savestate_header.magic_string, "TMSS", 4))
    goto invalid;

  // Right now this only supports savestate version 1. Later if more
  // are done it'll probably support those too! Version 0 is right out
  // though.

  if(savestate_header.version > TEMPER_SAVESTATE_VERSION)
    goto invalid;

  // Does it have a snapshot? Load it.
  if(savestate_header.extensions & SS_EXT_SNAPSHOT)
  {
    file_read_mem_load_fixed(savestate_file, 320 * 240 * 2);

    // Okay, must support these are they're added.
    switch(savestate_header.snapshot_format)
    {
      default:
      case SS_SNAPSHOT_FULL:
        file_read_mem(savestate_file, buffer, 320 * 240 * 2);
        break;
    }
  }

  free(savestate_buffer);
  return savestate_header.extensions;

 invalid:
  file_read_mem_close(savestate_file);
  free(savestate_buffer);
  return SS_EXT_INVALID;
}

#define config_file_action(type, version_gate)                                \
  file_##type##_array(config_file, config.pad);                               \
  file_##type##_variable(config_file, config.show_fps);                       \
  file_##type##_variable(config_file, config.enable_sound);                   \
  file_##type##_variable(config_file, config.fast_forward);                   \
  file_##type##_variable(config_file, config.audio_output_frequency);         \
  file_##type##_variable(config_file, config.patch_idle_loops);               \
  file_##type##_variable(config_file, config.snapshot_format);                \
  file_##type##_variable(config_file, config.force_usa);                      \
  file_##type##_variable(config_file, config.clock_speed);                    \
  file_##type##_variable(config_file, config.ram_timings);                    \
  file_##type##_variable(config_file, config.gamma_percent);                  \
  if(version_gate >= 2)                                                       \
  {                                                                           \
    file_##type##_variable(config_file, config.six_button_pad);               \
    file_##type##_variable(config_file, config.cd_system_type);               \
    file_##type##_variable(config_file, config.bz2_savestates);               \
    file_##type##_variable(config_file, config.per_game_bram);                \
    if(version_gate >= 3)                                                     \
    {                                                                         \
      file_##type##_variable(config_file, config.sound_double);               \
      file_##type##_variable(config_file, config.scale_factor);               \
      file_##type##_variable(config_file, config.scale_width);                \
      file_##type##_variable(config_file, config.unlimit_sprites);            \
      file_##type##_variable(config_file, config.compatibility_mode);         \
    }                                                                         \
  }                                                                           \
                                                                              \

void save_config_file(char *file_name)
{
  char path[MAX_PATH];
  sprintf(path, "%s%cconfig%c%s", config.main_path, DIR_SEPARATOR_CHAR,
   DIR_SEPARATOR_CHAR, file_name);
  u8 *config_file_buffer = malloc(16384);

  printf("saving config to file named %s\n", path);

  file_write_mem_open(config_file, path, config_file_buffer);
  config_header_struct config_header;

  // The header is TMCF for "Temper Config"
  config_header.magic_string[0] = 'T';
  config_header.magic_string[1] = 'M';
  config_header.magic_string[2] = 'C';
  config_header.magic_string[3] = 'F';

  config_header.version = TEMPER_CONFIG_VERSION;

  file_write_mem_array(config_file, config_header.magic_string);
  file_write_mem_variable(config_file, config_header.version);
  file_write_mem_array(config_file, config_header.reserved);

  config_file_action(write_mem, TEMPER_CONFIG_VERSION);

  file_write_mem_save(config_file);
  file_write_mem_close(config_file);
  free(config_file_buffer);
}

void save_directory_config_file(char *file_name)
{
  char path[MAX_PATH];
  sprintf(path, "%s%cconfig%c%s", config.main_path, DIR_SEPARATOR_CHAR,
   DIR_SEPARATOR_CHAR, file_name);
  u8 *config_file_buffer = malloc(16384);

  printf("saving directory config to file named %s\n", path);

  file_write_mem_open(config_file, path, config_file_buffer);
  config_header_struct config_header;

  // The header is TMC2 for "Temper Config type 2"
  config_header.magic_string[0] = 'T';
  config_header.magic_string[1] = 'M';
  config_header.magic_string[2] = 'C';
  config_header.magic_string[3] = '2';

  config_header.version = TEMPER_CONFIG2_VERSION;

  file_write_mem_array(config_file, config_header.magic_string);
  file_write_mem_variable(config_file, config_header.version);
  file_write_mem_array(config_file, config_header.reserved);

  file_write_mem_array(config_file, config.rom_directory);

  file_write_mem_save(config_file);
  file_write_mem_close(config_file);
  free(config_file_buffer);
}

s32 load_config_file(char *file_name)
{
  char path[MAX_PATH];
  sprintf(path, "%s%cconfig%c%s", config.main_path, DIR_SEPARATOR_CHAR,
   DIR_SEPARATOR_CHAR, file_name);

#ifdef CONFIG_OPTIONS_RAM_TIMINGS
  u32 old_ram_timings = config.ram_timings;
#endif

#ifdef CONFIG_OPTIONS_CLOCK_SPEED
  u32 old_clock_speed = config.clock_speed;
#endif

#ifdef CONFIG_OPTIONS_GAMMA
  u32 old_gamma_percent = config.gamma_percent;
#endif

  u32 i;

  config_buttons_enum v1_to_v2_button_config[] =
  {
    CONFIG_BUTTON_UP, CONFIG_BUTTON_DOWN, CONFIG_BUTTON_LEFT,
    CONFIG_BUTTON_RIGHT, CONFIG_BUTTON_I, CONFIG_BUTTON_II,
    CONFIG_BUTTON_RUN, CONFIG_BUTTON_SELECT, CONFIG_BUTTON_RAPID_I,
    CONFIG_BUTTON_RAPID_II, CONFIG_BUTTON_MENU, CONFIG_BUTTON_SAVE_STATE,
    CONFIG_BUTTON_LOAD_STATE, CONFIG_BUTTON_VOLUME_DOWN,
    CONFIG_BUTTON_VOLUME_UP, CONFIG_BUTTON_FAST_FORWARD, CONFIG_BUTTON_NONE
  };

  printf("loading config file %s\n", path);

  file_read_open(config_file, path);
  config_header_struct config_header;

  if(file_read_invalid(config_file))
  {
    printf("config file %s does not exist\n", path);
    return -1;
  }

  file_read_array(config_file, config_header.magic_string);
  file_read_variable(config_file, config_header.version);
  file_read_array(config_file, config_header.reserved);

  // First, the magic string must match.
  if(strncmp(config_header.magic_string, "TMCF", 4))
    goto invalid;

  if(config_header.version > TEMPER_CONFIG_VERSION)
    goto invalid;

  config_file_action(read, config_header.version);

  // The key map was changed in v2.
  if(config_header.version == 1)
  {
    for(i = 0; i < 16; i++)
    {
      config.pad[i] = v1_to_v2_button_config[config.pad[i]];
    }
  }

  // Others not allowed on the menu for now.
  config.audio_output_frequency = 44100;
  file_read_close(config_file);

#ifdef CONFIG_OPTIONS_RAM_TIMINGS
  if(config.ram_timings != old_ram_timings)
  {
    if(config.ram_timings)
      set_fast_ram_timings();
    else
      set_default_ram_timings();
  }
#endif

#ifdef CONFIG_OPTIONS_CLOCK_SPEED
  if(config.clock_speed != old_clock_speed)
    set_clock_speed(config.clock_speed);
#endif

#ifdef CONFIG_OPTIONS_GAMMA
  if(config.gp2x_gamma_percent != old_gamma_percent)
    set_gamma(config.gamma_percent);
#endif

  return 0;

 invalid:
  printf("an error occured while trying to load the config file\n");
  file_read_close(config_file);
  return -1;
}

s32 load_directory_config_file(char *file_name)
{
  char path[MAX_PATH];
  sprintf(path, "%s%cconfig%c%s", config.main_path, DIR_SEPARATOR_CHAR,
   DIR_SEPARATOR_CHAR, file_name);

  printf("loading directory config file %s\n", path);

  file_read_open(config_file, path);
  config_header_struct config_header;

  if(file_read_invalid(config_file))
  {
    printf("directory config file %s does not exist\n", path);
    return -1;
  }

  file_read_array(config_file, config_header.magic_string);
  file_read_variable(config_file, config_header.version);
  file_read_array(config_file, config_header.reserved);

  // First, the magic string must match.
  if(strncmp(config_header.magic_string, "TMC2", 4))
    goto invalid;

  // Current version is 1
  if(config_header.version > TEMPER_CONFIG2_VERSION)
    goto invalid;

  file_read_array(config_file, config.rom_directory);
  file_read_close(config_file);

  chdir(config.rom_directory);

  return 0;

 invalid:
  printf("an error occured while trying to load the directory config file\n");
  file_read_close(config_file);
  return -1;
}

s32 save_screenshot(u16 *snapshot)
{
  char images_path[MAX_PATH];
  char image_filename[MAX_PATH];
  char current_dir[MAX_PATH];
  sprintf(images_path, "%s%cimages", config.main_path, DIR_SEPARATOR_CHAR);
  u32 shot_number = 0;
  u16 bmp_header[54 / 2];
  u32 x, y;
  u16 current_pixel;
  u16 *bmp_offset;
  u8 *screenshot_buffer = malloc((320 * 240 * 2) + (26 * 2));

  struct stat sb;

  getcwd(current_dir, MAX_PATH);
  chdir(images_path);

  do
  {
    sprintf(image_filename, "%s_%d.bmp", config.rom_filename, shot_number);
    shot_number++;
  } while(!stat(image_filename, &sb));

  printf("saving image file %s\n", image_filename);

  file_write_mem_open(bmp_file, image_filename, screenshot_buffer);

  if(file_write_mem_invalid(bmp_file))
  {
    printf("could not create image file %s\n", image_filename);
    chdir(current_dir);
    return -1;
  }

  // 0 2  signature, must be 4D42 hex
  bmp_header[0] = 0x4D42;
  // 2 4 size of BMP file in bytes (unreliable)
  bmp_header[1] = ((320 * 240 * 2) + 54) & 0xFFFF;
  bmp_header[2] = ((320 * 240 * 2) + 54) >> 16;
  // 6 2 reserved, must be zero
  // 8 2 reserved, must be zero
  bmp_header[3] = 0;
  bmp_header[4] = 0;
  // 10 4 offset to start of image data in bytes
  bmp_header[5] = 54;
  bmp_header[6] = 0;
  // 14 4 size of BITMAPINFOHEADER structure, must be 40
  bmp_header[7] = 40;
  bmp_header[8] = 0;
  // 18 4 image width in pixels
  bmp_header[9] = 320;
  bmp_header[10] = 0;
  // 22 4 image height in pixels
  bmp_header[11] = 240;
  bmp_header[12] = 0;
  // 26 2 number of planes in the image, must be 1
  bmp_header[13] = 1;
  // 28 2 number of bits per pixel (1, 4, 8, 16, or 24)
  bmp_header[14] = 16;
  // 30 4 compression type (0=none, 1=RLE-8, 2=RLE-4)
  bmp_header[15] = 0;
  bmp_header[16] = 0;
  // 34 4 size of image data in bytes (including padding)
  bmp_header[17] = (320 * 240 * 2) & 0xFFFF;
  bmp_header[18] = (320 * 240 * 2) >> 16;
  // 38 4 horizontal resolution in pixels per meter (unreliable)
  bmp_header[19] = 0;
  bmp_header[20] = 0;
  // 42 4 vertical resolution in pixels per meter (unreliable)
  bmp_header[21] = 0;
  bmp_header[22] = 0;
  // 46 4 number of colors in image, or zero
  bmp_header[23] = 512;
  bmp_header[24] = 0;
  // 50 4 number of important colors, or zero
  bmp_header[25] = 512;
  bmp_header[26] = 0;

  file_write_mem_array(bmp_file, bmp_header);

  for(y = 0; y < 240; y++)
  {
    bmp_offset = snapshot + ((239 - y) * 320);

    for(x = 0; x < 320; x++)
    {
      current_pixel = bmp_offset[x];
      current_pixel = (current_pixel & 0x1F) |
       ((current_pixel >> 1) & ((0x1F << 5) | (0x1F << 10)));
      file_write_mem_variable(bmp_file, current_pixel);
    }
  }

  file_write_mem_save(bmp_file);
  file_write_mem_close(bmp_file);
  chdir(current_dir);
  free(screenshot_buffer);
  return 0;
}

void quit()
{
  char bram_path[MAX_PATH];
  get_bram_path(bram_path);
  save_bram(bram_path);

  chdir(config.main_path);

  save_directory_config_file("temper.cf2");
  audio_exit();

  platform_quit();

  exit(0);
}

