// Temper menu code

#include "common.h"

typedef enum
{
  MENU_OPTION_FIXED,
  MENU_OPTION_NUMERIC,
  MENU_OPTION_STRING
} menu_option_type_enum;

typedef struct
{
  struct _menu_struct *current_menu;
  u16 *snapshot_bg;
  u16 *current_bg;
  u32 bg_modified;
  u32 exit_menu;
  u32 column_start;
  u32 audio_rate_number;
  u32 restore_screen;
} menu_state_struct;

typedef struct
{
  char *name;
  u32 line_number;
  void (* menu_function)(menu_state_struct *menu_state);
  menu_option_type_enum menu_option_type;
  u32 *value;
  u32 values;
  char **value_strings;
  void (* modify_function)(menu_state_struct *menu_state);
} menu_option_struct;

struct _menu_struct
{
  u32 num_options;
  u32 current_option;
  void (* exit_function)(menu_state_struct *menu_state);
  menu_option_struct menu_options[];
};

typedef struct _menu_struct menu_struct;

#define count_menu_options(which)                                             \
  (sizeof(which) / sizeof(menu_option_struct))                                \


#define menu_option_fixed(name, line, function)                               \
  { name, line, function, MENU_OPTION_FIXED, NULL, 0, NULL, NULL }            \

#define menu_option_numeric(name, line, function, value, values)              \
  { name, line, function, MENU_OPTION_NUMERIC, (u32 *)&(value), values, NULL, \
  NULL }                                                                      \

#define menu_option_string(name, line, function, value, values, strs)         \
  { name, line, function, MENU_OPTION_STRING, (u32 *)&(value), values, strs,  \
  NULL }                                                                      \

#define menu_option_numeric_mod(name, line, function, value, values, mod)     \
  { name, line, function, MENU_OPTION_NUMERIC, (u32 *)&(value), values, NULL, \
  mod }                                                                       \

#define menu_option_string_mod(name, line, func, value, values, strs, mod)    \
  { name, line, func, MENU_OPTION_STRING, (u32 *)&(value), values, strs,      \
  mod }                                                                       \


#define button_count 25

char *button_strings[] =
{
  "          Up", "        Down", "        Left", "       Right",
  "           I", "          II", "         III", "          IV",
  "           V", "          VI", "         Run", "      Select",
  "     Rapid I", "    Rapid II", "   Rapid III", "    Rapid IV",
  "     Rapid V", "    Rapid VI", "        Menu", "  Save state",
  "  Load state", " Volume down", "   Volume up", "Fast forward",
  "        None"
};

char *yes_no_strings[] =
{
  " no", "yes"
};

char *audio_rate_strings[] =
{
  " 8000KHz", "11025KHz", "16000KHz", "22050KHz", "44100KHz"
};

char *cd_system_version_strings[] =
{
  "  v1", "  v2", "  v3", " acd", "gecd"
};


void menu_pad_done(menu_state_struct *menu_state);
void menu_pad_save_config_local(menu_state_struct *menu_state);
void menu_pad_save_config_global(menu_state_struct *menu_state);

void menu_options_done(menu_state_struct *menu_state);
void menu_options_save_config_local(menu_state_struct *menu_state);
void menu_options_save_config_global(menu_state_struct *menu_state);

menu_state_struct menu_state;



#ifdef GP2X_BUILD

#define CONFIG_EXIT_BUTTON          "x"
#define CONTROL_CONFIG_START_COLUMN 40
#define CONTROL_CONFIG_STRING       "GP2X                  PC-Engine"

menu_struct options_menu =
{
  19, 0, menu_options_done,
  {
    menu_option_string("Show fps                %s", 7, NULL,
     config.show_fps, 2, yes_no_strings),
    menu_option_string("Enable sound            %s", 8, NULL,
     config.enable_sound, 2, yes_no_strings),
    menu_option_string("Fast forward            %s", 9, NULL,
     config.fast_forward, 2, yes_no_strings),
    menu_option_numeric("Clock speed             %03d", 10, NULL,
     config.clock_speed, 301),
    menu_option_string("Fast RAM timings        %s", 11, NULL,
     config.ram_timings, 2, yes_no_strings),
    menu_option_numeric("Gamma percent           %03d", 12, NULL,
     config.gamma_percent, 300),
    menu_option_string("Patch idle loops        %s", 13, NULL,
     config.patch_idle_loops, 2, yes_no_strings),
    menu_option_string("Snapshot in saves       %s", 14, NULL,
     config.snapshot_format, 2, yes_no_strings),
    menu_option_string("BZ2 compressed saves    %s", 15, NULL,
     config.bz2_savestates, 2, yes_no_strings),
    menu_option_string("Use six button pad      %s", 16, NULL,
     config.six_button_pad, 2, yes_no_strings),
    menu_option_string("CD-ROM System          %s", 17, NULL,
     config.cd_system_type, 5, cd_system_version_strings),
    menu_option_string("Per-game BRAM saves     %s", 18, NULL,
     config.per_game_bram, 2, yes_no_strings),
    menu_option_string("Double volume           %s", 19, NULL,
     config.sound_double, 2, yes_no_strings),
    menu_option_string("Scale screen width      %s", 20, NULL,
     config.scale_width, 2, yes_no_strings),
    menu_option_string("Don't limit sprites     %s", 21, NULL,
     config.unlimit_sprites, 2, yes_no_strings),
    menu_option_string("Use compatibility mode  %s", 22, NULL,
     config.compatibility_mode, 2, yes_no_strings),
    menu_option_fixed("Exit: save for all games", 24,
     menu_options_save_config_global),
    menu_option_fixed("Exit: save for this game", 25,
     menu_options_save_config_local),
    menu_option_fixed("Exit without saving", 26, menu_options_done)
  }
};

menu_struct pad_menu =
{
  19, 0, menu_pad_done,
  {
    menu_option_string("Up                %s", 6, NULL, config.pad[0],
     button_count, button_strings),
    menu_option_string("Down              %s", 7, NULL, config.pad[1],
     button_count, button_strings),
    menu_option_string("Left              %s", 8, NULL, config.pad[2],
     button_count, button_strings),
    menu_option_string("Right             %s", 9, NULL, config.pad[3],
     button_count, button_strings),
    menu_option_string("A                 %s", 10, NULL, config.pad[4],
     button_count, button_strings),
    menu_option_string("B                 %s", 11, NULL, config.pad[5],
     button_count, button_strings),
    menu_option_string("X                 %s", 12, NULL, config.pad[6],
     button_count, button_strings),
    menu_option_string("Y                 %s", 13, NULL, config.pad[7],
     button_count, button_strings),
    menu_option_string("L Trigger         %s", 14, NULL, config.pad[8],
     button_count, button_strings),
    menu_option_string("R Trigger         %s", 15, NULL, config.pad[9],
     button_count, button_strings),
    menu_option_string("Start             %s", 16, NULL, config.pad[10],
     button_count, button_strings),
    menu_option_string("Select            %s", 17, NULL, config.pad[11],
     button_count, button_strings),
    menu_option_string("Volume Down       %s", 18, NULL, config.pad[12],
     button_count, button_strings),
    menu_option_string("Volume Up         %s", 19, NULL, config.pad[13],
     button_count, button_strings),
    menu_option_string("Volume Up + Down  %s", 20, NULL, config.pad[14],
     button_count, button_strings),
    menu_option_string("Stick Click       %s", 21, NULL, config.pad[15],
     button_count, button_strings),

    menu_option_fixed("Exit: save for all games", 23,
     menu_pad_save_config_global),
    menu_option_fixed("Exit: save for this game", 24,
     menu_pad_save_config_local),
    menu_option_fixed("Exit without saving", 25, menu_pad_done)
  }
};

#endif


#ifdef WIZ_BUILD

#define CONFIG_EXIT_BUTTON          "x"
#define CONTROL_CONFIG_START_COLUMN 40
#define CONTROL_CONFIG_STRING       "Wiz                  PC-Engine"

menu_struct options_menu =
{
  17, 0, menu_options_done,
  {
    menu_option_string("Show fps                %s", 7, NULL,
     config.show_fps, 2, yes_no_strings),
    menu_option_string("Enable sound            %s", 8, NULL,
     config.enable_sound, 2, yes_no_strings),
    menu_option_string("Fast forward            %s", 9, NULL,
     config.fast_forward, 2, yes_no_strings),
    menu_option_numeric("Clock speed             %03d", 10, NULL,
     config.clock_speed, 900),
    menu_option_string("Patch idle loops        %s", 11, NULL,
     config.patch_idle_loops, 2, yes_no_strings),
    menu_option_string("Snapshot in saves       %s", 12, NULL,
     config.snapshot_format, 2, yes_no_strings),
    menu_option_string("BZ2 compressed saves    %s", 13, NULL,
     config.bz2_savestates, 2, yes_no_strings),
    menu_option_string("Use six button pad      %s", 14, NULL,
     config.six_button_pad, 2, yes_no_strings),
    menu_option_string("CD-ROM System          %s", 15, NULL,
     config.cd_system_type, 5, cd_system_version_strings),
    menu_option_string("Per-game BRAM saves     %s", 16, NULL,
     config.per_game_bram, 2, yes_no_strings),
    menu_option_string("Double volume           %s", 17, NULL,
     config.sound_double, 2, yes_no_strings),
    menu_option_string("Scale screen width      %s", 18, NULL,
     config.scale_width, 2, yes_no_strings),
    menu_option_string("Don't limit sprites     %s", 19, NULL,
     config.unlimit_sprites, 2, yes_no_strings),
    menu_option_string("Use compatibility mode  %s", 20, NULL,
     config.compatibility_mode, 2, yes_no_strings),
    menu_option_fixed("Exit: save for all games", 22,
     menu_options_save_config_global),
    menu_option_fixed("Exit: save for this game", 23,
     menu_options_save_config_local),
    menu_option_fixed("Exit without saving", 24, menu_options_done)
  }
};

menu_struct pad_menu =
{
  18, 0, menu_pad_done,
  {
    menu_option_string("Up                %s", 6, NULL, config.pad[0],
     button_count, button_strings),
    menu_option_string("Down              %s", 7, NULL, config.pad[1],
     button_count, button_strings),
    menu_option_string("Left              %s", 8, NULL, config.pad[2],
     button_count, button_strings),
    menu_option_string("Right             %s", 9, NULL, config.pad[3],
     button_count, button_strings),
    menu_option_string("A                 %s", 10, NULL, config.pad[4],
     button_count, button_strings),
    menu_option_string("B                 %s", 11, NULL, config.pad[5],
     button_count, button_strings),
    menu_option_string("X                 %s", 12, NULL, config.pad[6],
     button_count, button_strings),
    menu_option_string("Y                 %s", 13, NULL, config.pad[7],
     button_count, button_strings),
    menu_option_string("L Trigger         %s", 14, NULL, config.pad[8],
     button_count, button_strings),
    menu_option_string("R Trigger         %s", 15, NULL, config.pad[9],
     button_count, button_strings),
    menu_option_string("Menu              %s", 16, NULL, config.pad[10],
     button_count, button_strings),
    menu_option_string("Select            %s", 17, NULL, config.pad[11],
     button_count, button_strings),
    menu_option_string("Volume Down       %s", 18, NULL, config.pad[12],
     button_count, button_strings),
    menu_option_string("Volume Up         %s", 19, NULL, config.pad[13],
     button_count, button_strings),
    menu_option_string("Volume Up + Down  %s", 20, NULL, config.pad[14],
     button_count, button_strings),

    menu_option_fixed("Exit: save for all games", 22,
     menu_pad_save_config_global),
    menu_option_fixed("Exit: save for this game", 23,
     menu_pad_save_config_local),
    menu_option_fixed("Exit without saving", 24, menu_pad_done)
  }
};

#endif


#ifdef PC_BUILD

#define CONFIG_EXIT_BUTTON          "escape"
#define CONTROL_CONFIG_START_COLUMN 64
#define CONTROL_CONFIG_STRING       "PC            PC-Engine"

char *scale_strings[] =
{
  "fullscreen", "        1x", "        2x", "        3x", "        4x"
};

menu_struct options_menu =
{
  16, 0, menu_options_done,
  {
    menu_option_string("Scale screen     %s", 8, NULL,
     config.scale_factor, 5, scale_strings),
    menu_option_string("Show fps                %s", 9, NULL,
     config.show_fps, 2, yes_no_strings),
    menu_option_string("Enable sound            %s", 10, NULL,
     config.enable_sound, 2, yes_no_strings),
    menu_option_string("Fast forward            %s", 11, NULL,
     config.fast_forward, 2, yes_no_strings),
    menu_option_string("Fast RAM timings        %s", 12, NULL,
     config.ram_timings, 2, yes_no_strings),
    menu_option_string("Patch idle loops        %s", 13, NULL,
     config.patch_idle_loops, 2, yes_no_strings),
    menu_option_string("Snapshot in saves       %s", 14, NULL,
     config.snapshot_format, 2, yes_no_strings),
    menu_option_string("BZ2 compressed saves    %s", 15, NULL,
     config.bz2_savestates, 2, yes_no_strings),
    menu_option_string("Use six button pad      %s", 16, NULL,
     config.six_button_pad, 2, yes_no_strings),
    menu_option_string("CD-ROM System          %s", 17, NULL,
     config.cd_system_type, 5, cd_system_version_strings),
    menu_option_string("Per-game BRAM saves     %s", 18, NULL,
     config.per_game_bram, 2, yes_no_strings),
    menu_option_string("Scale screen width      %s", 19, NULL,
     config.scale_width, 2, yes_no_strings),
    menu_option_string("Allow 17+ spr per line  %s", 20, NULL,
     config.unlimit_sprites, 2, yes_no_strings),
    menu_option_fixed("Exit: save for all games", 22,
     menu_options_save_config_global),
    menu_option_fixed("Exit: save for this game", 23,
     menu_options_save_config_local),
    menu_option_fixed("Exit without saving", 24, menu_options_done)
  }
};

// These are taken from my Logitech Rumblepad 2

menu_struct pad_menu =
{
  19, 0, menu_pad_done,
  {
    menu_option_string("Up         %s", 6, NULL, config.pad[0],
     button_count, button_strings),
    menu_option_string("Down       %s", 7, NULL, config.pad[1],
     button_count, button_strings),
    menu_option_string("Left       %s", 8, NULL, config.pad[2],
     button_count, button_strings),
    menu_option_string("Right      %s", 9, NULL, config.pad[3],
     button_count, button_strings),
    menu_option_string("1          %s", 10, NULL, config.pad[4],
     button_count, button_strings),
    menu_option_string("2          %s", 11, NULL, config.pad[5],
     button_count, button_strings),
    menu_option_string("3          %s", 12, NULL, config.pad[6],
     button_count, button_strings),
    menu_option_string("4          %s", 13, NULL, config.pad[7],
     button_count, button_strings),
    menu_option_string("5          %s", 14, NULL, config.pad[8],
     button_count, button_strings),
    menu_option_string("6          %s", 15, NULL, config.pad[9],
     button_count, button_strings),
    menu_option_string("7          %s", 16, NULL, config.pad[10],
     button_count, button_strings),
    menu_option_string("8          %s", 17, NULL, config.pad[11],
     button_count, button_strings),
    menu_option_string("9          %s", 18, NULL, config.pad[12],
     button_count, button_strings),
    menu_option_string("10         %s", 19, NULL, config.pad[13],
     button_count, button_strings),
    menu_option_string("11         %s", 20, NULL, config.pad[14],
     button_count, button_strings),
    menu_option_string("12         %s", 21, NULL, config.pad[15],
     button_count, button_strings),

    menu_option_fixed("Exit: save for all games", 23,
     menu_pad_save_config_global),
    menu_option_fixed("Exit: save for this game", 24,
     menu_pad_save_config_local),
    menu_option_fixed("Exit without saving", 25, menu_pad_done)
  }
};

#endif


#ifdef PSP_BUILD

#define CONFIG_EXIT_BUTTON          "cross"
#define CONTROL_CONFIG_START_COLUMN 50
#define CONTROL_CONFIG_STRING       "PSP                  PC-Engine"

char *psp_scale_strings[] =
{
  "none             ", "full aspect (4:3)", "widescreen       "
};

menu_struct options_menu =
{
  18, 0, menu_options_done,
  {
    menu_option_string("Scale screen            %s", 6, NULL,
     config.scale_factor, 5, psp_scale_strings),
    menu_option_string("Show fps                %s", 7, NULL,
     config.show_fps, 2, yes_no_strings),
    menu_option_string("Enable sound            %s", 8, NULL,
     config.enable_sound, 2, yes_no_strings),
    menu_option_string("Fast forward            %s", 9, NULL,
     config.fast_forward, 2, yes_no_strings),
    menu_option_numeric("Clock speed             %03d", 10, NULL,
     config.clock_speed, 334),
    menu_option_string("Patch idle loops        %s", 11, NULL,
     config.patch_idle_loops, 2, yes_no_strings),
    menu_option_string("Snapshot in saves       %s", 12, NULL,
     config.snapshot_format, 2, yes_no_strings),
    menu_option_string("BZ2 compressed saves    %s", 13, NULL,
     config.bz2_savestates, 2, yes_no_strings),
    menu_option_string("Use six button pad      %s", 14, NULL,
     config.six_button_pad, 2, yes_no_strings),
    menu_option_string("CD-ROM System          %s", 15, NULL,
     config.cd_system_type, 5, cd_system_version_strings),
    menu_option_string("Per-game BRAM saves     %s", 16, NULL,
     config.per_game_bram, 2, yes_no_strings),
    menu_option_string("Double volume           %s", 17, NULL,
     config.sound_double, 2, yes_no_strings),
    menu_option_string("Scale screen width      %s", 18, NULL,
     config.scale_width, 2, yes_no_strings),
    menu_option_string("Don't limit sprites     %s", 19, NULL,
     config.unlimit_sprites, 2, yes_no_strings),
    menu_option_string("Use compatibility mode  %s", 20, NULL,
     config.compatibility_mode, 2, yes_no_strings),
    menu_option_fixed("Exit: save for all games", 22,
     menu_options_save_config_global),
    menu_option_fixed("Exit: save for this game", 23,
     menu_options_save_config_local),
    menu_option_fixed("Exit without saving", 24, menu_options_done)
  }
};

menu_struct pad_menu =
{
  15, 0, menu_pad_done,
  {
    menu_option_string("Up         %s", 8, NULL, config.pad[0],
     button_count, button_strings),
    menu_option_string("Down       %s", 9, NULL, config.pad[1],
     button_count, button_strings),
    menu_option_string("Left       %s", 10, NULL, config.pad[2],
     button_count, button_strings),
    menu_option_string("Right      %s", 12, NULL, config.pad[3],
     button_count, button_strings),
    menu_option_string("Square     %s", 13, NULL, config.pad[4],
     button_count, button_strings),
    menu_option_string("Circle     %s", 14, NULL, config.pad[5],
     button_count, button_strings),
    menu_option_string("Cross      %s", 15, NULL, config.pad[6],
     button_count, button_strings),
    menu_option_string("Triangle   %s", 16, NULL, config.pad[7],
     button_count, button_strings),
    menu_option_string("L          %s", 17, NULL, config.pad[8],
     button_count, button_strings),
    menu_option_string("R          %s", 18, NULL, config.pad[9],
     button_count, button_strings),
    menu_option_string("Start      %s", 19, NULL, config.pad[10],
     button_count, button_strings),
    menu_option_string("Select     %s", 20, NULL, config.pad[11],
     button_count, button_strings),

    menu_option_fixed("Exit: save for all games", 22,
     menu_pad_save_config_global),
    menu_option_fixed("Exit: save for this game", 23,
     menu_pad_save_config_local),
    menu_option_fixed("Exit without saving", 24, menu_pad_done)
  }
};

#endif


#define menu_line(line)                                                       \
  (line * 8)                                                                  \

gui_action_type gui_wait_for_input()
{
  gui_action_type gui_action = CURSOR_NONE;

  do
  {
    gui_action = get_gui_input();
  } while(gui_action == CURSOR_NONE);

  return gui_action;
}

#define FILE_LIST_ROWS ((int)((240 - 40) / 10))
#define FILE_LIST_CHARS 36
#define DIR_LIST_CHARS 16
#define FILE_LIST_POSITION 6
#define DIR_LIST_POSITION (FILE_LIST_POSITION + (FILE_LIST_CHARS * 6) + 1)


int sort_function(const void *dest_str_ptr, const void *src_str_ptr)
{
  char *dest_str = *((char **)dest_str_ptr);
  char *src_str = *((char **)src_str_ptr);

  if(src_str[0] == '.')
    return 1;

  if(dest_str[0] == '.')
    return -1;

  return strcasecmp(dest_str, src_str);
}

#define move_cursor_down()                                                    \
  if(current_column == 0)                                                     \
  {                                                                           \
    if(current_file_selection < (num_files - 1))                              \
    {                                                                         \
      current_file_selection++;                                               \
      if(current_file_in_scroll == (FILE_LIST_ROWS - 1))                      \
      {                                                                       \
        do_screen_blit = 1;                                                   \
        current_file_scroll_value++;                                          \
      }                                                                       \
      else                                                                    \
      {                                                                       \
        current_file_in_scroll++;                                             \
      }                                                                       \
    }                                                                         \
  }                                                                           \
  else                                                                        \
  {                                                                           \
    if(current_dir_selection < (num_dirs - 1))                                \
    {                                                                         \
      current_dir_selection++;                                                \
      if(current_dir_in_scroll == (FILE_LIST_ROWS - 1))                       \
      {                                                                       \
        do_screen_blit = 1;                                                   \
        current_dir_scroll_value++;                                           \
      }                                                                       \
      else                                                                    \
      {                                                                       \
        current_dir_in_scroll++;                                              \
      }                                                                       \
    }                                                                         \
  }                                                                           \

#define move_cursor_up()                                                      \
  if(current_column == 0)                                                     \
  {                                                                           \
    if(current_file_selection)                                                \
    {                                                                         \
      current_file_selection--;                                               \
      if(current_file_in_scroll == 0)                                         \
      {                                                                       \
        do_screen_blit = 1;                                                   \
        current_file_scroll_value--;                                          \
      }                                                                       \
      else                                                                    \
      {                                                                       \
        current_file_in_scroll--;                                             \
      }                                                                       \
    }                                                                         \
  }                                                                           \
  else                                                                        \
  {                                                                           \
    if(current_dir_selection)                                                 \
    {                                                                         \
      current_dir_selection--;                                                \
      if(current_dir_in_scroll == 0)                                          \
      {                                                                       \
        do_screen_blit = 1;                                                   \
        current_dir_scroll_value--;                                           \
      }                                                                       \
      else                                                                    \
      {                                                                       \
        current_dir_in_scroll--;                                              \
      }                                                                       \
    }                                                                         \
  }                                                                           \

const u32 cursor_page_rate = FILE_LIST_ROWS;

s32 load_file(char **wildcards, char *result, u16 *screen_bg)
{
  DIR *current_dir;
  struct dirent *current_file;
  struct stat file_info;
  char current_dir_name[MAX_PATH];
  char current_dir_short[81];
  u32 current_dir_length;
  u32 total_filenames_allocated;
  u32 total_dirnames_allocated;
  char **file_list;
  char **dir_list;
  u32 num_files;
  u32 num_dirs;
  char *file_name;
  u32 file_name_length;
  u32 ext_pos = -1;
  u32 chosen_file, chosen_dir;
  s32 return_value = 1;
  u32 current_file_selection;
  u32 current_file_scroll_value;
  u32 current_dir_selection;
  u32 current_dir_scroll_value;
  u32 current_file_in_scroll;
  u32 current_dir_in_scroll;
  u32 current_file_number, current_dir_number;
  u32 current_column = 0;
  u32 repeat;
  u32 i;
  gui_action_type gui_action;
  u32 do_screen_blit;

  char name_buffer[FILE_LIST_CHARS + 1];
  char *name_buffer_ptr;

  blit_screen(screen_bg);
  set_font_narrow();

  while(return_value == 1)
  {
    current_file_selection = 0;
    current_file_scroll_value = 0;
    current_dir_selection = 0;
    current_dir_scroll_value = 0;
    current_file_in_scroll = 0;
    current_dir_in_scroll = 0;

    total_filenames_allocated = 32;
    total_dirnames_allocated = 32;
    file_list = malloc(sizeof(char *) * 32);
    dir_list = malloc(sizeof(char *) * 32);
    memset(file_list, 0, sizeof(char *) * 32);
    memset(dir_list, 0, sizeof(char *) * 32);

    num_files = 0;
    num_dirs = 0;
    chosen_file = 0;
    chosen_dir = 0;

    getcwd(current_dir_name, MAX_PATH);

    current_dir = opendir(current_dir_name);

    do
    {
      if(current_dir)
        current_file = readdir(current_dir);
      else
        current_file = NULL;

      if(current_file)
      {
        file_name = current_file->d_name;
        file_name_length = strlen(file_name);

        if((stat(file_name, &file_info) >= 0) &&
         ((file_name[0] != '.') || (file_name[1] == '.')))
        {
          if(S_ISDIR(file_info.st_mode))
          {
            dir_list[num_dirs] =
             malloc(file_name_length + 1);

            sprintf(dir_list[num_dirs], "%s", file_name);

            num_dirs++;
          }
          else
          {
            // Must match one of the wildcards, also ignore the .
            if(file_name_length >= 4)
            {
              if(file_name[file_name_length - 4] == '.')
                ext_pos = file_name_length - 4;
              else

              if(file_name[file_name_length - 3] == '.')
                ext_pos = file_name_length - 3;

              else
                ext_pos = 0;

              for(i = 0; wildcards[i] != NULL; i++)
              {
                if(!strcasecmp((file_name + ext_pos),
                 wildcards[i]))
                {
                  file_list[num_files] =
                   malloc(file_name_length + 1);

                  sprintf(file_list[num_files], "%s", file_name);

                  num_files++;
                  break;
                }
              }
            }
          }
        }

        if(num_files == total_filenames_allocated)
        {
          file_list = realloc(file_list, sizeof(char *) *
           total_filenames_allocated * 2);
          memset(file_list + total_filenames_allocated, 0,
           sizeof(char *) * total_filenames_allocated);
          total_filenames_allocated *= 2;
        }

        if(num_dirs == total_dirnames_allocated)
        {
          dir_list = realloc(dir_list, sizeof(char *) *
           total_dirnames_allocated * 2);
          memset(dir_list + total_dirnames_allocated, 0,
           sizeof(char *) * total_dirnames_allocated);
          total_dirnames_allocated *= 2;
        }
      }
    } while(current_file);

    qsort((void *)file_list, num_files, sizeof(char *), sort_function);
    qsort((void *)dir_list, num_dirs, sizeof(char *), sort_function);

    closedir(current_dir);

    current_dir_length = strlen(current_dir_name);

    if(current_dir_length > 80)
    {
      snprintf(current_dir_short, 80,
       "...%s", current_dir_name + current_dir_length - 77);
    }
    else
    {
      snprintf(current_dir_short, 80, "%s", current_dir_name);
    }

    repeat = 1;

    if(num_files == 0)
      current_column = 1;

    while(repeat)
    {
      print_string_bg(current_dir_short, 0xFFFF, screen_bg, 6, 4);
      print_string_bg("Press " CONFIG_EXIT_BUTTON
       " to return to the main menu.", 0xFFFF, screen_bg, 20, 228);

      for(i = 0, current_file_number = i + current_file_scroll_value;
       i < FILE_LIST_ROWS; i++, current_file_number++)
      {
        if(current_file_number < num_files)
        {
          name_buffer_ptr = file_list[current_file_number];
          if(strlen(name_buffer_ptr) >= FILE_LIST_CHARS)
          {
            memcpy(name_buffer, name_buffer_ptr, FILE_LIST_CHARS);
            name_buffer_ptr = name_buffer;
            name_buffer_ptr[FILE_LIST_CHARS] = 0;
          }

          if((current_file_number == current_file_selection) &&
           (current_column == 0))
          {
            print_string(name_buffer_ptr, 0xFFFF, 0x0017,
             FILE_LIST_POSITION, (i * 10) + 20);
          }
          else
          {
            print_string_bg(name_buffer_ptr, 0xFFFF, screen_bg,
             FILE_LIST_POSITION, (i * 10) + 20);
          }
        }
      }

      for(i = 0, current_dir_number = i + current_dir_scroll_value;
       i < FILE_LIST_ROWS; i++, current_dir_number++)
      {
        if(current_dir_number < num_dirs)
        {
          name_buffer_ptr = dir_list[current_dir_number];
          if(strlen(name_buffer_ptr) >= DIR_LIST_CHARS)
          {
            memcpy(name_buffer, name_buffer_ptr, DIR_LIST_CHARS);
            name_buffer_ptr = name_buffer;
            name_buffer_ptr[DIR_LIST_CHARS] = 0;
          }

          if((current_dir_number == current_dir_selection) &&
           (current_column == 1))
          {
            print_string(name_buffer_ptr, 0xFFFF,
             0x0017, DIR_LIST_POSITION, (i * 10) + 20);
          }
          else
          {
            print_string_bg(name_buffer_ptr, 0xFFFF,
             screen_bg, DIR_LIST_POSITION, (i * 10) + 20);
          }
        }
      }

      update_screen();

      gui_action = gui_wait_for_input();
      do_screen_blit = 0;

      switch(gui_action)
      {
        case CURSOR_DOWN:
          move_cursor_down();
          break;

        case CURSOR_UP:
          move_cursor_up();
          break;

        // Cheap but effective.
        case CURSOR_PAGE_UP:
          for(i = 0; i < cursor_page_rate; i++)
          {
            move_cursor_up();
          }
          break;

        case CURSOR_PAGE_DOWN:
          for(i = 0; i < cursor_page_rate; i++)
          {
            move_cursor_down();
          }
          break;

        case CURSOR_RIGHT:
          if(current_column == 0)
          {
            if(num_dirs != 0)
              current_column = 1;
          }
          break;

        case CURSOR_LEFT:
          if(current_column == 1)
          {
            if(num_files != 0)
              current_column = 0;
          }
          break;

        case CURSOR_SELECT:
          if(current_column == 1)
          {
            repeat = 0;
            chdir(dir_list[current_dir_selection]);
            do_screen_blit = 1;
          }
          else
          {
            if(num_files != 0)
            {
              repeat = 0;
              return_value = 0;
              strcpy(result, file_list[current_file_selection]);
            }
          }
          break;

        case CURSOR_BACK:
          repeat = 0;
          chdir("..");
          do_screen_blit = 1;
          break;

        case CURSOR_EXIT:
          return_value = -1;
          repeat = 0;
          break;

        default:
          break;
      }

      if(do_screen_blit)
        blit_screen(screen_bg);
    }

    for(i = 0; i < num_files; i++)
    {
      free(file_list[i]);
    }
    free(file_list);

    for(i = 0; i < num_dirs; i++)
    {
      free(dir_list[i]);
    }
    free(dir_list);
  }

  return return_value;
}

char *pce_ext[] = { ".pce", ".bz2", ".cue", NULL };


void menu_draw_main(menu_state_struct *menu_state)
{
  menu_state->column_start = 104;
  blit_screen(menu_state->current_bg);

  set_font_wide();

  print_string_bg("Temper version " TEMPER_VERSION, 0xFFFF,
   menu_state->current_bg, 88, menu_line(4));
}

void menu_load_game(menu_state_struct *menu_state)
{
  char file_loaded[MAX_PATH];

  if(load_file(pce_ext, file_loaded, menu_state->current_bg) != -1)
  {
    if(load_rom(file_loaded) != -1)
    {
      reset_pce();
      menu_state->exit_menu = 1;
      menu_state->restore_screen = 0;
    }
    else
    {
      menu_draw_main(menu_state);
    }
  }
  else
  {
    menu_draw_main(menu_state);
  }
}

char *cd_ext[] = { ".cue", NULL };

void menu_swap_cd(menu_state_struct *menu_state)
{
  char file_loaded[MAX_PATH];
  if(load_file(cd_ext, file_loaded, menu_state->current_bg) != -1)
  {
    load_bin_cue(file_loaded);
    menu_state->exit_menu = 1;
    menu_state->restore_screen = 0;
  }
  else
  {
    menu_draw_main(menu_state);
  }
}

void menu_update_savestate(menu_state_struct *menu_state)
{
  if(config.rom_filename[0])
  {
    char state_name[MAX_PATH];
    sprintf(state_name, "%s_%d.svs", config.rom_filename,
     config.savestate_number);

    if(load_state_snapshot(state_name, menu_state->current_bg) &
     SS_EXT_SNAPSHOT)
    {
      menu_state->bg_modified = 1;
      buffer_half_intensity(menu_state->current_bg,
       menu_state->current_bg, 320);
      menu_draw_main(menu_state);
    }
    else

    if(menu_state->bg_modified)
    {
      menu_state->bg_modified = 0;
      buffer_quarter_intensity(menu_state->current_bg,
       menu_state->snapshot_bg, 320);
      menu_draw_main(menu_state);
    }
  }
}

void menu_pad(menu_state_struct *menu_state)
{
  blit_screen(menu_state->current_bg);

  menu_state->column_start = CONTROL_CONFIG_START_COLUMN;
  print_string_bg(CONTROL_CONFIG_STRING, 0xFFFF,
   menu_state->current_bg, menu_state->column_start, menu_line(4));

  menu_state->current_menu = &pad_menu;
}

u32 old_clock_speed;
u32 old_ram_timings;
u32 old_gamma_percent;
u32 old_scale_width;

void menu_options(menu_state_struct *menu_state)
{
  blit_screen(menu_state->current_bg);

  old_ram_timings = config.ram_timings;
  old_clock_speed = config.clock_speed;
  old_gamma_percent = config.gamma_percent;
  old_scale_width = config.scale_width;

  menu_state->column_start = 56;
  print_string_bg("Configure Options", 0xFFFF,
   menu_state->current_bg, menu_state->column_start, menu_line(7));

  menu_state->current_menu = &options_menu;
}

void menu_load_state(menu_state_struct *menu_state)
{
  if(config.rom_filename[0])
  {
    char state_name[MAX_PATH];
    sprintf(state_name, "%s_%d.svs", config.rom_filename,
     config.savestate_number);

    load_state(state_name);
    menu_state->exit_menu = 1;
  }
}

void menu_save_state(menu_state_struct *menu_state)
{
  if(config.rom_filename[0])
  {
    char state_name[MAX_PATH];
    sprintf(state_name, "%s_%d.svs", config.rom_filename,
     config.savestate_number);

    if((config.snapshot_format) == SS_SNAPSHOT_OFF)
      save_state(state_name, NULL);
    else
      save_state(state_name, menu_state->snapshot_bg);
  }
}

void menu_save_snapshot(menu_state_struct *menu_state)
{
  if(config.rom_filename[0])
    save_screenshot(menu_state->snapshot_bg);
}

void menu_restart(menu_state_struct *menu_state)
{
  if(config.rom_filename[0])
  {
    reset_pce();
    menu_state->exit_menu = 1;
    menu_state->restore_screen = 0;
  }
}

void menu_quit(menu_state_struct *menu_state)
{
  quit();
}

void menu_return(menu_state_struct *menu_state)
{
  if(config.rom_filename[0])
    menu_state->exit_menu = 1;
}

#ifdef CD_SWAP_OPTION

menu_struct main_menu =
{
  10, 0, menu_return,
  {
    menu_option_fixed("Change options", 13, menu_options),
    menu_option_fixed("Configure pad ", 14, menu_pad),
    menu_option_numeric_mod("Load state  %d ", 16, menu_load_state,
     config.savestate_number, 10, menu_update_savestate),
    menu_option_numeric("Save state  %d ", 17, menu_save_state,
     config.savestate_number, 10),
    menu_option_fixed("Save snapshot ", 18, menu_save_snapshot),
    menu_option_fixed("Load new game ", 20, menu_load_game),
    menu_option_fixed("Restart game  ", 21, menu_restart),
    menu_option_fixed("Swap CD       ", 22, menu_swap_cd),
    menu_option_fixed("Return to game", 24, menu_return),
    menu_option_fixed("Exit Temper   ", 26, menu_quit)
  }
};

#else

menu_struct main_menu =
{
  9, 0, menu_return,
  {
    menu_option_fixed("Change options", 14, menu_options),
    menu_option_fixed("Configure pad ", 15, menu_pad),
    menu_option_numeric_mod("Load state  %d ", 17, menu_load_state,
     config.savestate_number, 10, menu_update_savestate),
    menu_option_numeric("Save state  %d ", 18, menu_save_state,
     config.savestate_number, 10),
    menu_option_fixed("Save snapshot ", 19, menu_save_snapshot),
    menu_option_fixed("Load new game ", 21, menu_load_game),
    menu_option_fixed("Restart game  ", 22, menu_restart),
    menu_option_fixed("Return to game", 23, menu_return),
    menu_option_fixed("Exit Temper   ", 25, menu_quit)
  }
};

#endif

u32 audio_rate_number_frequencies[] = { 8000, 11025, 16000, 22050, 44100 };

void menu_options_done(menu_state_struct *menu_state)
{
  config.audio_output_frequency =
   audio_rate_number_frequencies[menu_state->audio_rate_number];

#ifdef CONFIG_OPTIONS_CLOCK_SPEED
  if(old_clock_speed != config.clock_speed)
    set_clock_speed(config.clock_speed);
#endif

#ifdef CONFIG_OPTIONS_RAM_TIMINGS
  if(old_ram_timings != config.ram_timings)
  {
    if(config.ram_timings)
      set_ram_timings_fast();
    else
      set_ram_timings_default()
  }
#endif

#ifdef CONFIG_OPTIONS_GAMMA
  if(old_gamma_percent != config.gamma_percent)
    set_gamma(config.gamma_percent);
#endif

  if(old_scale_width > config.scale_width)
  {
    u32 i;
    for(i = 0; i < 240; i++)
    {
      vdc.scanline_widths[i] = 320;
    }
  }

  menu_draw_main(menu_state);
  menu_state->current_menu = &main_menu;
}

void menu_pad_done(menu_state_struct *menu_state)
{
#if(defined(GP2X_BUILD) || defined(PSP_BUILD))
  u32 i;

  // If nothing's mapped to the menu we have a problem. Better force it
  // for something.

  // This is only an issue for the non-PC version, because it has the
  // M key to fall back on.

  for(i = 0; i < 16; i++)
  {
    if(config.pad[i] == CONFIG_BUTTON_MENU)
      break;
  }

  if(i == 16)
#ifdef GP2X_BUILD
    config.pad[GP2X_CONF_VOL_MID] = CONFIG_BUTTON_MENU;
#else
    config.pad[PSP_CONF_TRIANGLE] = CONFIG_BUTTON_MENU;
#endif

#endif

  menu_draw_main(menu_state);
  menu_state->current_menu = &main_menu;
}

void save_config_local(menu_state_struct *menu_state)
{
  if(config.rom_filename[0])
  {
    char config_file_name[MAX_PATH];
    sprintf(config_file_name, "%s.cfg", config.rom_filename);
    save_config_file(config_file_name);
  }
}

void save_config_global(menu_state_struct *menu_state)
{
  save_config_file("temper.cfg");
}


void menu_options_save_config_local(menu_state_struct *menu_state)
{
  menu_options_done(menu_state);
  save_config_local(menu_state);
}

void menu_options_save_config_global(menu_state_struct *menu_state)
{
  menu_options_done(menu_state);
  save_config_global(menu_state);
}

void menu_pad_save_config_local(menu_state_struct *menu_state)
{
  menu_pad_done(menu_state);
  save_config_local(menu_state);
}

void menu_pad_save_config_global(menu_state_struct *menu_state)
{
  menu_pad_done(menu_state);
  save_config_global(menu_state);
}


void menu(u32 start_file_dialog)
{
  u32 i;

  u16 *screen_capture = malloc(320 * 240 * sizeof(u16));
  u16 *screen_bg = malloc(320 * 240 * sizeof(u16));

  char menu_option[128];
  char *menu_option_ptr;

  gui_action_type current_action;

  menu_struct *current_menu = &main_menu;
  menu_option_struct *current_menu_option;

  void (* last_modify_function)(menu_state_struct *menu_state);

  menu_state.bg_modified = 0;
  menu_state.current_menu = current_menu;
  menu_state.snapshot_bg = screen_capture;
  menu_state.current_bg = screen_bg;
  menu_state.exit_menu = 0;
  menu_state.restore_screen = 1;

  switch(config.audio_output_frequency)
  {
    case 8000:
      menu_state.audio_rate_number = 0;
      break;

    case 11025:
      menu_state.audio_rate_number = 1;
      break;
      break;

    case 16000:
      menu_state.audio_rate_number = 2;
      break;
      break;

    case 22050:
      menu_state.audio_rate_number = 3;
      break;
      break;

    case 44100:
      menu_state.audio_rate_number = 4;
      break;
      break;
  }

  set_single_buffer_mode();

  audio_pause();

  copy_screen(screen_capture);
  copy_screen_quarter_intensity(screen_bg);

  if(start_file_dialog)
    menu_load_game(&menu_state);

  menu_draw_main(&menu_state);

  current_menu_option = current_menu->menu_options +
   current_menu->current_option;
  last_modify_function = current_menu_option->modify_function;

  while(menu_state.exit_menu == 0)
  {
    current_menu = menu_state.current_menu;

    current_menu_option = current_menu->menu_options +
     current_menu->current_option;

    if(current_menu_option->modify_function)
    {
      current_menu_option->modify_function(&menu_state);
    }
    else

    if((last_modify_function != NULL) && menu_state.bg_modified)
    {
      menu_state.current_bg = screen_bg;
      buffer_quarter_intensity(menu_state.current_bg,
       screen_capture, 320);
      menu_draw_main(&menu_state);
      menu_state.bg_modified = 0;
    }

    for(i = 0; i < current_menu->num_options; i++)
    {
      current_menu_option = current_menu->menu_options + i;

      menu_option_ptr = current_menu_option->name;

      switch(current_menu_option->menu_option_type)
      {
        case MENU_OPTION_FIXED:
          break;

        case MENU_OPTION_NUMERIC:
          sprintf(menu_option, menu_option_ptr,
           *(current_menu_option->value));
          menu_option_ptr = menu_option;
          break;

        case MENU_OPTION_STRING:
          sprintf(menu_option, menu_option_ptr,
           current_menu_option->value_strings[*(current_menu_option->value)]);
          menu_option_ptr = menu_option;
          break;
      }

      if(i == current_menu->current_option)
      {
        print_string(menu_option_ptr, 0xFFFF, 0x0017,
         menu_state.column_start, current_menu_option->line_number * 8);
      }
      else
      {
        print_string_bg(menu_option_ptr, 0xFFFF, screen_bg,
         menu_state.column_start, current_menu_option->line_number * 8);
      }
    }

    current_menu_option =
     current_menu->menu_options + current_menu->current_option;

    update_screen();

    current_action = gui_wait_for_input();

    last_modify_function = current_menu_option->modify_function;

    switch(current_action)
    {
      case CURSOR_UP:
        if(current_menu->current_option == 0)
          current_menu->current_option = current_menu->num_options - 1;
        else
          current_menu->current_option = current_menu->current_option - 1;
        break;

      case CURSOR_DOWN:
        if(current_menu->current_option == current_menu->num_options - 1)
          current_menu->current_option = 0;
        else
          current_menu->current_option = current_menu->current_option + 1;
        break;

      case CURSOR_LEFT:
        if(current_menu_option->menu_option_type != MENU_OPTION_FIXED)
        {
          if(*(current_menu_option->value) == 0)
            *(current_menu_option->value) = current_menu_option->values - 1;
          else
            (*(current_menu_option->value))--;
        }
        break;

      case CURSOR_RIGHT:
        if(current_menu_option->menu_option_type != MENU_OPTION_FIXED)
        {
          (*(current_menu_option->value))++;

          if(*(current_menu_option->value) == current_menu_option->values)
            *(current_menu_option->value) = 0;
        }
        break;

      case CURSOR_SELECT:
        if(current_menu_option->menu_function != NULL)
          current_menu_option->menu_function(&menu_state);

        break;

      case CURSOR_EXIT:
        if(current_menu->exit_function != NULL)
          current_menu->exit_function(&menu_state);
        break;

      default:
        break;
    }
  }

  audio_unpause();

  // Wait for buttons to be released.
  while(get_gui_input() != CURSOR_NONE);
  clear_gui_actions();

  set_multi_buffer_mode();
  clear_all_buffers();

  if(menu_state.restore_screen)
    blit_screen(menu_state.snapshot_bg);

  free(screen_capture);
  free(screen_bg);
}

